# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""
The :class:`Response` class models the response from your application to a
single request.
"""
from __future__ import absolute_import

import re
import json
from itertools import chain

import fresco
from .cookie import Cookie
from .compat import ustr

__all__ = [
    'STATUS_CONTINUE', 'STATUS_SWITCHING_PROTOCOLS', 'STATUS_PROCESSING',
    'STATUS_OK', 'STATUS_CREATED', 'STATUS_ACCEPTED',
    'STATUS_NON_AUTHORITATIVE_INFORMATION', 'STATUS_NO_CONTENT',
    'STATUS_RESET_CONTENT', 'STATUS_PARTIAL_CONTENT', 'STATUS_MULTI_STATUS',
    'STATUS_ALREADY_REPORTED', 'STATUS_IM_USED', 'STATUS_MULTIPLE_CHOICES',
    'STATUS_MOVED_PERMANENTLY', 'STATUS_FOUND', 'STATUS_SEE_OTHER',
    'STATUS_NOT_MODIFIED', 'STATUS_USE_PROXY', 'STATUS_TEMPORARY_REDIRECT',
    'STATUS_PERMANENT_REDIRECT', 'STATUS_BAD_REQUEST', 'STATUS_UNAUTHORIZED',
    'STATUS_PAYMENT_REQUIRED', 'STATUS_FORBIDDEN', 'STATUS_NOT_FOUND',
    'STATUS_METHOD_NOT_ALLOWED', 'STATUS_NOT_ACCEPTABLE',
    'STATUS_PROXY_AUTHENTICATION_REQUIRED', 'STATUS_REQUEST_TIMEOUT',
    'STATUS_CONFLICT', 'STATUS_GONE', 'STATUS_LENGTH_REQUIRED',
    'STATUS_PRECONDITION_FAILED', 'STATUS_PAYLOAD_TOO_LARGE',
    'STATUS_URI_TOO_LONG', 'STATUS_UNSUPPORTED_MEDIA_TYPE',
    'STATUS_RANGE_NOT_SATISFIABLE', 'STATUS_EXPECTATION_FAILED',
    'STATUS_UNPROCESSABLE_ENTITY', 'STATUS_LOCKED', 'STATUS_FAILED_DEPENDENCY',
    'STATUS_UPGRADE_REQUIRED', 'STATUS_PRECONDITION_REQUIRED',
    'STATUS_TOO_MANY_REQUESTS', 'STATUS_REQUEST_HEADER_FIELDS_TOO_LARGE',
    'STATUS_INTERNAL_SERVER_ERROR', 'STATUS_NOT_IMPLEMENTED',
    'STATUS_BAD_GATEWAY', 'STATUS_SERVICE_UNAVAILABLE',
    'STATUS_GATEWAY_TIMEOUT', 'STATUS_HTTP_VERSION_NOT_SUPPORTED',
    'STATUS_VARIANT_ALSO_NEGOTIATES', 'STATUS_INSUFFICIENT_STORAGE',
    'STATUS_LOOP_DETECTED', 'STATUS_NOT_EXTENDED',
    'STATUS_NETWORK_AUTHENTICATION_REQUIRED',
    'Response'
]


#: HTTP/1.1 status codes as listed in http://www.ietf.org/rfc/rfc2616.txt
HTTP_STATUS_CODES = {100: 'Continue',
                     101: 'Switching Protocols',
                     102: 'Processing',
                     200: 'OK',
                     201: 'Created',
                     202: 'Accepted',
                     203: 'Non-Authoritative Information',
                     204: 'No Content',
                     205: 'Reset Content',
                     206: 'Partial Content',
                     207: 'Multi-Status',
                     208: 'Already Reported',
                     226: 'IM Used',
                     300: 'Multiple Choices',
                     301: 'Moved Permanently',
                     302: 'Found',
                     303: 'See Other',
                     304: 'Not Modified',
                     305: 'Use Proxy',
                     307: 'Temporary Redirect',
                     308: 'Permanent Redirect',
                     400: 'Bad Request',
                     401: 'Unauthorized',
                     402: 'Payment Required',
                     403: 'Forbidden',
                     404: 'Not Found',
                     405: 'Method Not Allowed',
                     406: 'Not Acceptable',
                     407: 'Proxy Authentication Required',
                     408: 'Request Timeout',
                     409: 'Conflict',
                     410: 'Gone',
                     411: 'Length Required',
                     412: 'Precondition Failed',
                     413: 'Payload Too Large',
                     414: 'URI Too Long',
                     415: 'Unsupported Media Type',
                     416: 'Range Not Satisfiable',
                     417: 'Expectation Failed',
                     422: 'Unprocessable Entity',
                     423: 'Locked',
                     424: 'Failed Dependency',
                     426: 'Upgrade Required',
                     428: 'Precondition Required',
                     429: 'Too Many Requests',
                     431: 'Request Header Fields Too Large',
                     500: 'Internal Server Error',
                     501: 'Not Implemented',
                     502: 'Bad Gateway',
                     503: 'Service Unavailable',
                     504: 'Gateway Timeout',
                     505: 'HTTP Version Not Supported',
                     506: 'Variant Also Negotiates',
                     507: 'Insufficient Storage',
                     508: 'Loop Detected',
                     510: 'Not Extended',
                     511: 'Network Authentication Required'}

# Symbolic names for the HTTP status codes
STATUS_CONTINUE = 100
STATUS_SWITCHING_PROTOCOLS = 101
STATUS_PROCESSING = 102
STATUS_OK = 200
STATUS_CREATED = 201
STATUS_ACCEPTED = 202
STATUS_NON_AUTHORITATIVE_INFORMATION = 203
STATUS_NO_CONTENT = 204
STATUS_RESET_CONTENT = 205
STATUS_PARTIAL_CONTENT = 206
STATUS_MULTI_STATUS = 207
STATUS_ALREADY_REPORTED = 208
STATUS_IM_USED = 226
STATUS_MULTIPLE_CHOICES = 300
STATUS_MOVED_PERMANENTLY = 301
STATUS_FOUND = 302
STATUS_SEE_OTHER = 303
STATUS_NOT_MODIFIED = 304
STATUS_USE_PROXY = 305
STATUS_TEMPORARY_REDIRECT = 307
STATUS_PERMANENT_REDIRECT = 308
STATUS_BAD_REQUEST = 400
STATUS_UNAUTHORIZED = 401
STATUS_PAYMENT_REQUIRED = 402
STATUS_FORBIDDEN = 403
STATUS_NOT_FOUND = 404
STATUS_METHOD_NOT_ALLOWED = 405
STATUS_NOT_ACCEPTABLE = 406
STATUS_PROXY_AUTHENTICATION_REQUIRED = 407
STATUS_REQUEST_TIMEOUT = 408
STATUS_CONFLICT = 409
STATUS_GONE = 410
STATUS_LENGTH_REQUIRED = 411
STATUS_PRECONDITION_FAILED = 412
STATUS_PAYLOAD_TOO_LARGE = 413
STATUS_URI_TOO_LONG = 414
STATUS_UNSUPPORTED_MEDIA_TYPE = 415
STATUS_RANGE_NOT_SATISFIABLE = 416
STATUS_EXPECTATION_FAILED = 417
STATUS_UNPROCESSABLE_ENTITY = 422
STATUS_LOCKED = 423
STATUS_FAILED_DEPENDENCY = 424
STATUS_UPGRADE_REQUIRED = 426
STATUS_PRECONDITION_REQUIRED = 428
STATUS_TOO_MANY_REQUESTS = 429
STATUS_REQUEST_HEADER_FIELDS_TOO_LARGE = 431
STATUS_INTERNAL_SERVER_ERROR = 500
STATUS_NOT_IMPLEMENTED = 501
STATUS_BAD_GATEWAY = 502
STATUS_SERVICE_UNAVAILABLE = 503
STATUS_GATEWAY_TIMEOUT = 504
STATUS_HTTP_VERSION_NOT_SUPPORTED = 505
STATUS_VARIANT_ALSO_NEGOTIATES = 506
STATUS_INSUFFICIENT_STORAGE = 507
STATUS_LOOP_DETECTED = 508
STATUS_NOT_EXTENDED = 510
STATUS_NETWORK_AUTHENTICATION_REQUIRED = 511

#: Mapping from python symbolic names to HTTP headers to ensure headers are
#: emitted with correct capitalization
HEADER_NAMES = {
    'accept_ranges': 'Accept-Ranges',
    'age': 'Age',
    'allow': 'Allow',
    'cache_control': 'Cache-Control',
    'connection': 'Connection',
    'content_encoding': 'Content-Encoding',
    'content_language': 'Content-Language',
    'content_length': 'Content-Length',
    'content_location': 'Content-Location',
    'content_md5': 'Content-MD5',
    'content_disposition': 'Content-Disposition',
    'content_range': 'Content-Range',
    'content_type': 'Content-Type',
    'date': 'Date',
    'etag': 'ETag',
    'expires': 'Expires',
    'last_modified': 'Last-Modified',
    'link': 'Link',
    'location': 'Location',
    'p3p': 'P3P',
    'pragma': 'Pragma',
    'proxy_authenticate': 'Proxy-Authenticate',
    'refresh': 'Refresh',
    'retry_after': 'Retry-After',
    'server': 'Server',
    'set_cookie': 'Set-Cookie',
    'strict_transport_security': 'Strict-Transport-Security',
    'trailer': 'Trailer',
    'transfer_encoding': 'Transfer-Encoding',
    'vary': 'Vary',
    'via': 'Via',
    'warning': 'Warning',
    'www_authenticate': 'WWW-Authenticate',
    'x_frame_options': 'X-Frame-Options',
    'x_content_type_options': 'X-Content-Type-Options',
    'x_forwarded_proto': 'X-Forwarded-Proto',
    'front_end_https': 'Front-End-Https',
    'x_powered_by': 'X-Powered-By',
    'x_ua_compatible': 'X-UA-Compatible',
}

default_charset = 'UTF-8'


def encoder(stream, charset):
    r"""\
    Encode a response iterator using the given character set.
    """
    if charset is None:
        charset = default_charset

    for chunk in stream:
        if not isinstance(chunk, bytes):
            yield chunk.encode(charset)
        else:
            yield chunk


class Response(object):
    """\
    Model an HTTP response
    """

    default_content_type = "text/html; charset=UTF-8"

    def __init__(self, content=None, status=None, headers=None,
                 onclose=None, _nocontent=[], passthrough=False, **kwargs):
        """
        Create a new Response object, modelling the HTTP status, headers and
        content of an HTTP response. Response instances are valid WSGI
        applications.

        :param content: The response content as an iterable object
        :param status: The HTTP status line, eg ``200 OK`` or ``404 Not Found``
        :param headers: A list of HTTP headers
        :param passthrough: If True, use the content iterator unmodified.
                            Default behaviour is to wrap the content iterator
                            with :func:`encoder` to encode unicode strings
                            before output.
        :param kwargs: Arbitrary headers, provided as keyword arguments.
                       Underscores will be replaced with hyphens (eg
                       ``content_length`` becomes ``Content-Length``).

        Example usage::

            >>> # Construct a response
            >>> r = Response(
            ...     content=['hello world'],
            ...     status='200 OK',
            ...     headers=[('Content-Type', 'text/plain')]
            ... )
            >>>

        Changing headers or content::

            >>> r = r.add_header('X-Header', 'hello!')
            >>> r = r.replace(content=['whoa nelly!'],
            ...               content_type='text/html')

        """
        if content is None:
            content = _nocontent
            if status is None:
                status = '204 No Content'

        if status is None:
            status = '200 OK'

        self.status = make_status(status)
        if onclose is None:
            self.onclose = []
        elif callable(onclose):
            self.onclose = [onclose]
        else:
            self.onclose = list(onclose)

        if headers is None:
            headers = []
        headers = make_headers(headers, kwargs)

        # Ensure a content-type header is set if a content iterator has been
        # provided
        if content is not _nocontent:
            for key, value in headers:
                if key == 'Content-Type':
                    break
            else:
                headers.insert(0, ('Content-Type', self.default_content_type))

        self.headers = headers

        # Now we've dealt with the headers (including Content-Type,
        # which the charset property will look for when deciding how to
        # encode strings) we can pull the content into an iterable.
        # We optimize for the common cases (byte string, unicode string,
        # list of byte strings, list of unicode strings), otherwise
        # we wrap the iterator in :func:`encoder`, which iterates the content
        # and tries to encode each item as a byte string. Note that this
        # code assumes that lists are homogeneous.
        self.content = content
        content_islist = isinstance(content, list)
        if passthrough:
            self.content_iterator = content
        elif content_islist and len(content) == 0:
            self.content_iterator = content
        elif content_islist and isinstance(content[0], bytes):
            self.content_iterator = content
        elif content_islist and isinstance(content[0], ustr):
            charset = self.charset
            self.content_iterator = [c.encode(charset) for c in content]
        elif isinstance(content, bytes):
            self.content_iterator = [content]
        elif isinstance(content, ustr):
            self.content_iterator = [content.encode(self.charset)]
        else:
            self.content_iterator = _copy_close(
                content, encoder(content, self.charset))

    def __call__(self, environ, start_response, exc_info=None):
        """
        WSGI callable. Calls ``start_response`` with assigned headers and
        returns an iterator over ``content``.
        """
        start_response(self.status, self.headers, exc_info)
        result = self.content_iterator
        if self.onclose:
            result = ClosingIterator(result, *self.onclose)
        return result

    def add_onclose(self, *funcs):
        """
        Add functions to be called as part of the response iterator's ``close``
        method.
        """
        return self.__class__(self.content,
                              self.status,
                              self.headers,
                              self.onclose + list(funcs))

    @classmethod
    def from_wsgi(cls, wsgi_callable, environ, start_response):
        """
        Return a ``Response`` object constructed from the result of calling
        ``wsgi_callable`` with the given ``environ`` and ``start_response``
        arguments.
        """
        responder = StartResponseWrapper(start_response)
        content = wsgi_callable(environ, responder)
        if responder.buf.tell():
            content = _copy_close(content,
                                  chain(content, [responder.buf.getvalue()]))

        if not responder.called:
            # Iterator has not called start_response yet. Call next(content)
            # to force the application to call start_response
            try:
                chunk = next(content)
            except StopIteration:
                pass
            except Exception:
                close = getattr(content, 'close', None)
                if close is not None:
                    close()
                raise
            else:
                content = _copy_close(content, chain([chunk], content))
        return cls(content, responder.status, headers=responder.headers,
                   passthrough=True)

    def get_headers(self, name):
        """\
        Return the list of headers set with the given name.

        Synopsis::

            >>> r = Response(set_cookie = ['cookie1', 'cookie2'])
            >>> r.get_headers('set-cookie')
            ['cookie1', 'cookie2']

        """
        return [value for header, value in self.headers
                      if header.lower() == name.lower()]

    def get_header(self, name, default=''):
        """\
        Return the concatenated values of the named header(s) or ``default`` if
        the header has not been set.

        As specified in RFC2616 (section 4.2), multiple headers will be
        combined using a single comma.

        Example usage::

            >>> r = Response(set_cookie = ['cookie1', 'cookie2'])
            >>> r.get_header('set-cookie')
            'cookie1,cookie2'
        """
        headers = self.get_headers(name)
        if not headers:
            return default
        return ','.join(headers)

    @property
    def status_code(self):
        """
        Return the numeric status code for the response as an integer::

            >>> Response(status='404 Not Found').status_code
            404
            >>> Response(status=200).status_code
            200
        """
        return int(self.status.split(' ', 1)[0])

    @property
    def content_type(self):
        """\
        Return the value of the ``Content-Type`` header if set, otherwise
        ``None``.
        """
        for key, val in self.headers:
            if key.lower() == 'content-type':
                return val
        return None

    def add_header(self, name, value):
        """\
        Return a new response object with the given additional header.

        Synopsis::

            >>> r = Response(content_type='text/plain')
            >>> r.headers
            [('Content-Type', 'text/plain')]
            >>> r.add_header('Cache-Control', 'no-cache').headers
            [('Content-Type', 'text/plain'), ('Cache-Control', 'no-cache')]
        """
        return self.replace(headers=self.headers + [(name, value)])

    def add_headers(self, headers=[], **kwheaders):
        """\
        Return a new response object with the given additional headers.

        Synopsis::

            >>> r = Response(content_type='text/plain')
            >>> r.headers
            [('Content-Type', 'text/plain')]
            >>> r.add_headers(
            ...     cache_control='no-cache',
            ... ).headers
            [('Content-Type', 'text/plain'), ('Cache-Control', 'no-cache')]
        """
        return self.replace(headers=make_headers(self.headers + headers,
                                                      kwheaders))

    def remove_headers(self, *headers):
        """\
        Return a new response object with the named headers removed.

        Synopsis::

            >>> r = Response(content_type='text/plain',
            ...              cache_control='no-cache')
            >>> r.headers
            [('Cache-Control', 'no-cache'), ('Content-Type', 'text/plain')]
            >>> r.remove_headers('Cache-Control').headers
            [('Content-Type', 'text/plain')]
        """
        toremove = [item.lower() for item in headers]
        return self.replace(headers=[h for h in self.headers
                                       if h[0].lower() not in toremove])

    def add_cookie(self, name, value, max_age=None, expires=None, path='/',
                   secure=None, domain=None, comment=None, httponly=False):
        """\
        Return a new response object with the given cookie added.

        Synopsis::

            >>> r = Response(content_type='text/plain')
            >>> r.headers
            [('Content-Type', 'text/plain')]
            >>> r.add_cookie('a', '1').headers
            [('Content-Type', 'text/plain'), ('Set-Cookie', 'a=1;Version=1')]
        """
        return self.add_header('Set-Cookie', Cookie(name, value, max_age,
                                                    expires, path, secure,
                                                    domain,
                                                    comment=comment,
                                                    httponly=httponly))

    def delete_cookie(self, name, path='/', domain=None):
        return self.add_cookie(name, '', max_age=0, path=path, domain=domain,
                               expires='Thu, 01-Jan-1970 00:00:00 GMT')

    def add_vary(self, *vary_on):
        """
        Return a new response object with the given Vary header.
        Values specified will be added to any existing vary header.

        Synopsis::

            >>> r = Response().add_vary('Accept-Encoding')
            >>> r.headers
            [('Vary', 'Accept-Encoding')]
        """
        vary_on = set(vary_on)
        newheaders = []
        for k, v in self.headers:
            if k.lower() == 'vary':
                vary_on.update(s.strip() for s in v.split(','))
            else:
                newheaders.append((k, v))
        return self.replace(
            headers=newheaders + [('Vary', ', '.join(vary_on))])

    def replace(self, content=None, status=None, headers=None, **kwheaders):
        """\
        Return a new response object with any of content, status or headers
        changed.

        Synopsis::

            >>> r = Response(content_type='text/html')
            >>> r = r.replace(content='foo',
            ...               status=404,
            ...               headers=[('Content-Type', 'text/plain')],
            ...               content_length=3)

        """
        if content is None:
            content = self.content
            onclose = self.onclose
        else:
            close = getattr(self.content, 'close', None)
            onclose = self.onclose
            if close:
                onclose = [close] + onclose

        if headers is None:
            headers = self.headers

        if status is None:
            status = self.status

        if kwheaders:
            toremove = set(make_header_name(k) for k in kwheaders)
            headers = ([(k, v) for k, v in headers if k not in toremove] +
                       make_headers([], kwheaders))

        return self.__class__(content, status, headers, onclose=onclose)

    def buffered(self):
        """\
        Return a new response object with the content buffered into a list.
        This will also generate a content-length header.

        Example usage::

            >>> def generate_content():
            ...     yield "one two "
            ...     yield "three four five"
            ...
            >>> r = Response(content=generate_content())
            >>> r.content  # doctest: +ELLIPSIS
            <generator object ...>
            >>> r = Response(content=generate_content()).buffered()
            >>> r.content
            ['one two ', 'three four five']
        """
        content = list(self.content_iterator)
        content_length = sum(map(len, content))
        return self.replace(content=content, content_length=content_length)

    @property
    def charset(
        self,
        _parser=re.compile(r'.*;\s*charset=([\w\d\-]+)', re.I).match
    ):
        for key, val in self.headers:
            if key.lower() == 'content-type':
                mo = _parser(val)
                if mo:
                    return mo.group(1)
                break
        return default_charset

    @classmethod
    def not_found(cls, request=None):
        """\
        Return an HTTP not found response (404).

        Synopsis::

            >>> def view():
            ...     return Response.not_found()
            ...
        """
        return cls(
            status=STATUS_NOT_FOUND,
            content=[
                "<html>\n"
                "<body>\n"
                "    <h1>Not found</h1>\n"
                "    <p>The requested resource could not be found.</p>\n"
                "</body>\n"
                "</html>"
            ]
        )

    @classmethod
    def forbidden(cls, message='Sorry, access is denied'):
        """\
        Return an HTTP forbidden response (403).

        Synopsis::

            >>> def view():
            ...     return Response.forbidden()
            ...
        """
        return cls(
            "<html>\n"
            "<body>\n"
            "<h1>%s</h1>\n"
            "</body>\n"
            "</html>" % (message,),
            status=STATUS_FORBIDDEN,
        )

    @classmethod
    def bad_request(cls, request=None):
        """\
        Return an HTTP bad request response.

        Synopsis::

            >>> def view():
            ...     return Response.bad_request()
            ...

        """
        return cls(
            status=STATUS_BAD_REQUEST,
            content=[
                "<html>"
                "<body>"
                "<h1>The server could not understand your request</h1>"
                "</body>"
                "</html>"
            ]
        )

    @classmethod
    def length_required(cls, request=None):
        """\
        Return an HTTP Length Required response (411).

        Synopsis::

            >>> def view():
            ...     return Response.length_required()
            ...

        """
        return cls(
            status=STATUS_LENGTH_REQUIRED,
            content=[
                "<html>"
                "<body>"
                "<h1>A Content-Length header is required</h1>"
                "</body>"
                "</html>"
            ]
        )

    @classmethod
    def payload_too_large(cls, request=None):
        """\
        Return an HTTP Payload Too Large response (413)::

            >>> response = Response.payload_too_large()

        """
        return cls(
            status=STATUS_PAYLOAD_TOO_LARGE,
            content=["<html>"
                     "<body>"
                     "<h1>Payload Too Large</h1>"
                     "</body>"
                     "</html>"])

    request_entity_too_large = payload_too_large

    @classmethod
    def method_not_allowed(cls, valid_methods):
        """\
        Return an HTTP method not allowed response (405)::

            >>> from fresco import context
            >>> def view():
            ...     if context.request.method == 'POST':
            ...         return Response.method_not_allowed(('POST', ))
            ...

        :param valid_methods: A list of HTTP methods valid for requested URL

        :return: A :class:`fresco.response.Response` instance
        """

        return cls(
            status=STATUS_METHOD_NOT_ALLOWED,
            allow=",".join(valid_methods),
            content=["<html>"
                     "<body>"
                     "<h1>Method not allowed</h1>"
                     "</body>"
                     "</html>"
                     ])

    @classmethod
    def internal_server_error(cls):
        """\
        Return an HTTP internal server error response (500).

        Synopsis::

            >>> def view():
            ...     return Response.internal_server_error()
            ...

        :return: A :class:`fresco.response.Response` instance
        """

        return cls(
            status=STATUS_INTERNAL_SERVER_ERROR,
            content=["<html>"
                     "<body>"
                     "<h1>Internal Server Error</h1>"
                     "</body>"
                     "</html>"])

    @classmethod
    def redirect(cls, location, request=None, status=STATUS_FOUND, **kwargs):
        """\
        Return an HTTP redirect reponse (302 or 301).

        :param location: The redirect location or a view specification.
        :param status: HTTP status code for the redirect, default is
                       ``STATUS_FOUND`` (temporary redirect)

        Synopsis:

            >>> def view():
            ...   return Response.redirect("/new-location")
            ...

        The location parameter is interpreted as follows:

        - If it is a callable it is assumed to be a view,
          and passed to :func:`fresco.core.urlfor` to resolve it to a URL.

        - If it is a string and contains '://' it is assumed to be an absolute
          URL and no further processing is done.

        - If it is a string and contains a slash,
          it is assumed to be a
          relative URL and resolved relative to the current request.

        - If it is any other string
          it is passed to :func:`fresco.core.urlfor` to resolve it to a URL.
          If this fails (ie raises RouteNotFound) it is assumed to be a
          relative URL and resolved relative to the current request.
        """
        if callable(location):
            location = fresco.context.app.urlfor(location, **kwargs)

        elif '://' in location:
            pass

        elif '/' in location:
            if request is None:
                request = fresco.context.request
            location = request.resolve_url(location)

        else:
            from fresco.routing import RouteNotFound
            try:
                location = fresco.context.app.urlfor(location, **kwargs)
            except RouteNotFound:
                if kwargs:
                    # kwargs only makes sense if this is a view spec
                    raise
                if request is None:
                    request = fresco.context.request
                location = request.resolve_url(location)

        return Response(
            "<html><head></head><body>\n"
            "<h1>Page has moved</h1>\n"
            "<p><a href='%s'>%s</a></p>\n"
            "</body></html>" % (location, location),
            status=status,
            location=location)

    @classmethod
    def redirect_permanent(cls, *args, **kwargs):
        """\
        Return an HTTP permanent redirect reponse.

        :param location: the URI of the new location. If relative this will be
                         converted to an absolute URL based on the current
                         request.

        """
        return cls.redirect(status=STATUS_MOVED_PERMANENTLY, *args, **kwargs)

    @classmethod
    def redirect_temporary(cls, *args, **kwargs):
        """\
        Return an HTTP permanent redirect reponse.

        :param location: the URI of the new location. If relative this will be
                         converted to an absolute URL based on the current
                         request.

        """
        return cls.redirect(status=STATUS_FOUND, *args, **kwargs)

    @classmethod
    def meta_refresh(cls, location, delay=1, request=None):
        """\
        Return an HTML page containing a <meta http-equiv="refresh"> tag,
        causing the browser to redirect to the given location after ``delay``
        seconds.

        :param location: the URI of the new location. If relative this will be
                         converted to an absolute URL based on the current
                         request.

        """
        if '://' not in location:
            if request is None:
                request = fresco.context.request
            location = request.resolve_url(location)
        return cls([(
            '<!DOCTYPE html>'
            '<html>'
            '<head><meta http-equiv="refresh" content="0; url={0}"></head>'
            '<body><p><a href="{0}">Click here to continue</a></p></body>'
            '</html>').format(location)
        ], content_type='text/html')

    @classmethod
    def json(cls,
             data,
             indent=None,
             separators=(',', ':'),
             content_type='application/json',
             status=None,
             headers=None,
             dumps=json.dumps,
             **kwargs):
        """
        Return an ``application/json`` response with the given data
        JSON serialized

        :param data: The data to json encode.
        :param indent: The indent level.
                       Defaults to ``None`` (no pretty printing)
        :param separators: Defaults to ``(',', ':')`` for the most compact JSON
                           representation
        :param kwargs: Other keyword arguments are passed to ``json.dumps``.
                       These may be used to change encoding paramters, for
                       example overriding the default ``JSONEncoder`` class.

        """
        return cls([dumps(data,
                          indent=indent,
                          separators=separators,
                          **kwargs)],
                   status=status,
                   headers=headers,
                   content_type=content_type,)


def make_header_name(name):
    """\
    Return a formatted header name from a python idenfier.

    Example usage::

        >>> make_header_name('content_type')
        'Content-Type'
    """
    try:
        return HEADER_NAMES[name]
    except KeyError:
        return name.replace('_', '-').title()


def make_headers(header_list, header_dict):
    """
    Return a list of header (name, value) tuples from the combination of
    the header_list and header_dict.

    Synopsis::

        >>> make_headers(
        ...     [('Content-Type', 'text/html')],
        ...     {'content_length' : 54}
        ... )
        [('Content-Type', 'text/html'), ('Content-Length', '54')]

        >>> make_headers(
        ...     [('Content-Type', 'text/html')],
        ...     {'x_foo' : ['a1', 'b2']}
        ... )
        [('Content-Type', 'text/html'), ('X-Foo', 'a1'), ('X-Foo', 'b2')]

    """
    headers = header_list + list(header_dict.items())
    headers = [(make_header_name(key), val)
               for key, val in headers if val is not None]

    # Join multiple headers, see RFC2616 section 4.2
    newheaders = []
    for key, val in headers:
        if isinstance(val, list):
            for item in val:
                newheaders.append((key, str(item)))
        else:
            newheaders.append((key, str(val)))
    return newheaders


def dump_response(r, line_break=b'\r\n', encoding='UTF-8'):
    """\
    Return a byte-string representation of the given response, as for a HTTP
    response.
    """
    output = []
    output.append(r.status.encode('ascii'))
    output.append(line_break)
    for k, v in sorted(r.headers):
        output.append(k.encode('ascii'))
        output.append(b': ')
        output.append(v.encode('ascii'))
        output.append(line_break)
    output.append(line_break)
    output.extend(r.content_iterator)
    s = b''.join(output)
    if encoding:
        return s.decode(encoding)
    return s


def make_status(status):
    """
    Return a status line from the given status, which may be a simple integer.

    Synopsis::

        >>> make_status(200)
        '200 OK'

    """
    if isinstance(status, int):
        return '%d %s' % (status, HTTP_STATUS_CODES[status])
    return status


def _copy_close(src, dst, marker=object()):
    """\
    Copy the ``close`` attribute from ``src`` to ``dst``, which are assumed to
    be iterators.

    If it is not possible to copy the attribute over (eg for
    ``itertools.chain``, which does not support the close attribute) an
    instance of ``ClosingIterator`` is returned which will proxy calls to
    ``close`` as necessary.
    """

    close = getattr(src, 'close', marker)
    if close is not marker:
        try:
            setattr(dst, 'close', close)
        except AttributeError:
            return ClosingIterator(dst, close)

    return dst

from .util.wsgi import StartResponseWrapper, ClosingIterator



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from fresco.util.common import object_or_404
from fresco.exceptions import NotFound
import pytest


class TestObjectOr404(object):

    def test_it_returns_the_object(self):
        ob = object()
        assert object_or_404(ob) is ob

    def test_it_raises_NotFound(self):
        with pytest.raises(NotFound):
            object_or_404(None)

    def test_it_raises_custom_exception(self):
        class Foo(Exception):
            pass
        with pytest.raises(Foo):
            object_or_404(None, exception=Foo)



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

from wsgiref.util import setup_testing_defaults

from mock import Mock, call
import pytest

from fresco import Request, Response
from fresco import FrescoApp, GET, POST, PUT, DELETE
from fresco import Route, urlfor, context
from fresco.util.wsgi import ClosingIterator, make_environ
from . import fixtures


def create_env(**env):
    setup_testing_defaults(env)
    return env


class TestFrescoApp(object):

    def test_route_operates_as_a_decorator(self):

        app = FrescoApp()

        @app.route('/', GET)
        def view():
            return Response([b'ok'])

        with app.requestcontext('/'):
            assert list(app.view().content_iterator) == [b'ok']

    def test_route_operates_as_a_function(self):

        def view():
            return Response([b'ok'])

        app = FrescoApp()
        app.route('/', GET, view)
        with app.requestcontext('/'):
            assert list(app.view().content_iterator) == [b'ok']

    def test_route_returns_route_instance(self):
        def view():
            return Response([b'ok'])

        app = FrescoApp()
        assert isinstance(app.route('/', GET, view), Route)

    def test_route_http_methods(self):

        def view():
            return Response([context.request.environ['REQUEST_METHOD']])

        app = FrescoApp()
        app.route('/get', GET, view)
        app.route('/post', POST, view)

        with app.requestcontext('/get', REQUEST_METHOD='GET'):
            assert app.view().status_code == 200

        with app.requestcontext('/get', REQUEST_METHOD='POST'):
            assert app.view().status_code == 405

        with app.requestcontext('/post', REQUEST_METHOD='GET'):
            assert app.view().status_code == 405

        with app.requestcontext('/post', REQUEST_METHOD='POST'):
            assert app.view().status_code == 200

    def test_HEAD_request_delegated_to_GET_view(self):

        app = FrescoApp()

        @app.route('/', GET)
        def view():
            return Response([], x_original_view='GET')

        with app.requestcontext('/', REQUEST_METHOD='HEAD'):
            assert app.view().get_header('X-Original-View') == 'GET'

    def test_NotFound_observed_when_raised_in_handler(self):

        def app1():
            from fresco.exceptions import NotFound
            if 'foo' in context.request.path_info:
                raise NotFound()
            return Response([b'app1'])

        def app2():
            return Response([b'app2'])

        app = FrescoApp()
        app.route_all('/', GET, app1)
        app.route_all('/', GET, app2)

        with app.requestcontext('/bar'):
            assert list(app.view().content_iterator) == [b'app1']

        with app.requestcontext('/foo'):
            assert list(app.view().content_iterator) == [b'app2']

    def test_NotFound_final_observed_when_raised_in_handler(self):

        def app1():
            from fresco.exceptions import NotFound
            if 'foo' in context.request.path_info:
                raise NotFound(final=True)
            return Response([b'app1'])

        def app2():
            return Response([b'app2'])

        app = FrescoApp()
        app.route_all('/', GET, app1)
        app.route_all('/', GET, app2)

        with app.requestcontext('/bar'):
            assert list(app.view().content_iterator) == [b'app1']

        with app.requestcontext('/foo/'):
            assert app.view().status_code == 404

    def test_apps_called_in_correct_order(self):

        def view(value=''):
            return Response([value])

        app = FrescoApp()
        app.route_all('/f', GET, view, value=b'foo')
        app.route_all('/', GET, view, value=b'bar')

        with app.requestcontext('/f/bar'):
            assert list(app.view().content_iterator) == [b'foo']

        with app.requestcontext('/b/bar'):
            assert list(app.view().content_iterator) == [b'bar']

    def test_wsgi_app_handles_response_exceptions(self):

        from fresco.exceptions import NotFound

        def view():
            raise NotFound()

        app = FrescoApp()
        app.route('/', GET, view)

        with app.requestcontext('/'):
            assert app.view().status_code == 404

    def test_route_wsgi_app(self):

        def wsgiapp(environ, start_response):

            start_response('200 OK',
                           [('Content-Type', 'application/x-pachyderm')])
            return [b'pretty pink elephants']

        app = FrescoApp()
        app.route_wsgi('/', wsgiapp)

        with app.requestcontext('/'):
            assert list(app.view().content_iterator) == \
                    [b'pretty pink elephants']
            assert app.view().get_header('Content-Type') == \
                    'application/x-pachyderm'

    def test_get_methods_matches_on_path(self):

        app = FrescoApp()
        app.route('/1', POST, lambda: None)
        app.route('/1', PUT, lambda: None)
        app.route('/2', GET, lambda: None)
        app.route('/2', DELETE, lambda: None)

        with app.requestcontext() as c:
            assert app.get_methods(app, c.request, '/1') == set([POST, PUT])

        with app.requestcontext() as c:
            assert app.get_methods(app, c.request, '/2') == set([GET, DELETE])

    def test_get_methods_matches_on_predicate(self):

        p1 = Mock(return_value=True)
        p2 = Mock(return_value=False)

        app = FrescoApp()
        app.route('/', POST, lambda: None, predicate=p1)
        app.route('/', PUT, lambda: None, predicate=p2)

        with app.requestcontext('/') as c:
            assert app.get_methods(app, c.request, '/') == set([POST])
            assert p1.call_args_list == [call(c.request)]
            assert p2.call_args_list == [call(c.request)]

    def test_invalid_path_encoding_triggers_bad_request(self):
        app = FrescoApp()
        with app.requestcontext(
                PATH_INFO=fixtures.misquoted_wsgi_unicode_path):
            assert app.view().status_code == 400

    def test_remove_middleware(self):
        app = FrescoApp()
        m1 = Mock()
        m2 = Mock()
        app.add_middleware(m1)
        app.add_middleware(m2)
        app.remove_middleware(m1)
        app(create_env(), Mock()).close()
        assert m1.call_count == 0
        assert m2.call_count == 1

    def test_insert_middleware(self):

        calls = []

        def middleware(app, name):

            def middleware(env, start_response):
                calls.append(name)
                return app(env, start_response)
            return middleware

        app = FrescoApp()
        app.add_middleware(middleware, 'venus')
        app.add_middleware(middleware, 'mercury')
        app.insert_middleware(0, middleware, 'earth')

        # Put a request through app to initialize the WSGI stack
        with app.requestcontext('/'):
            app.view()

        # Middleware is called from the outside-in, so the item inserted in
        # position 0 is last to be called
        assert calls == ['mercury', 'venus', 'earth']


class TestAppHooks(object):

    def test_process_request_continues_on_none(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        app.route('/', GET, view)

        @app.process_request
        def process_request(request):
            assert isinstance(request, Request)
            return None

        with app.requestcontext('/'):
            app.view()
            assert view.call_count == 1

    def test_process_request_replaces_response(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        app.route('/', GET, view)
        new_response = Response()

        @app.process_request
        def process_request(request):
            return new_response

        with app.requestcontext('/'):
            assert app.view() is new_response
            assert view.call_count == 0

    def test_process_view_continues_on_none(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        app.route('/', GET, view)

        @app.process_view
        def process_view(request, v, args, kwargs):
            assert v is view
            return None

        with app.requestcontext('/'):
            app.view()
            assert view.call_count == 1

    def test_process_view_replaces_view(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        view2 = Mock(return_value=Response())
        app.route('/', GET, view)

        @app.process_view
        def process_view(request, v, args, kwargs):
            assert v is view
            return view2

        with app.requestcontext('/'):
            app.view()
            assert view.call_count == 0
            assert view2.call_count == 1

    def test_process_view_replaces_response(self):
        app = FrescoApp()
        new_response = Response()
        view = Mock(return_value=Response())
        app.route('/', GET, view)

        @app.process_view
        def process_view(request, v, args, kwargs):
            assert v is view
            return new_response

        with app.requestcontext('/'):
            assert app.view() is new_response
            assert view.call_count == 0

    def test_process_response_continues_on_none(self):
        app = FrescoApp()
        response = Response()
        view = Mock(return_value=response)
        app.route('/', GET, view)

        @app.process_response
        def process_response(req, res):
            assert res is response
            return None

        with app.requestcontext('/'):
            assert app.view() is response

    def test_process_response_replaces_response(self):
        app = FrescoApp()
        new_response = Response()
        view = Mock(return_value=Response())
        app.route('/', GET, view)

        @app.process_response
        def process_response(req, res):
            return new_response

        with app.requestcontext('/'):
            assert app.view() is new_response

    def test_process_http_error_response_continues_on_none(self):
        app = FrescoApp()
        response = Response.not_found()
        view = Mock(return_value=response)
        app.route('/', GET, view)

        @app.process_http_error_response
        def process_http_error_response(req, res):
            return None

        with app.requestcontext('/'):
            assert app.view() is response

    def test_process_http_error_response_replaces_response(self):
        app = FrescoApp()
        new_response = Response()
        view = Mock(return_value=Response.not_found())
        app.route('/', GET, view)

        @app.process_http_error_response
        def process_http_error_response(req, res):
            return new_response

        with app.requestcontext('/'):
            assert app.view() is new_response

    def test_process_http_error_response_called_on_internal_errors(self):
        app = FrescoApp()
        new_response = Response()
        view = Mock(return_value=Response())
        app.route('/', GET, view)

        hook = Mock(return_value=new_response)
        app.process_http_error_response(hook)

        # POSTing to a GET route should generate 405 method not supported
        with app.requestcontext_post('/'):
            assert app.view() is new_response
            hook_args, hook_kwargs = hook.call_args
            assert hook_args[1].status_code == 405

        # Should generate a 404 not found
        with app.requestcontext_post('/asdf'):
            assert app.view() is new_response
            hook_args, hook_kwargs = hook.call_args
            assert hook_args[1].status_code == 404

    def test_process_http_error_response_not_called_on_success(self):
        app = FrescoApp()
        ok_response = Response([''], status='200 OK')
        redirect_response = Response([''], status='302 Found', location='/')
        ok_view = Mock(return_value=ok_response)
        redirect_view = Mock(return_value=redirect_response)
        app.route('/ok', GET, ok_view)
        app.route('/redirect', GET, redirect_view)

        @app.process_http_error_response
        def process_http_error_response(req, res):
            assert False, "Hook should not be called"

        with app.requestcontext('/ok'):
            assert app.view() is ok_response

        with app.requestcontext('/redirect'):
            assert app.view() is redirect_response

    def test_process_http_error_response_can_be_associated_with_status(self):
        app = FrescoApp()
        process_404 = Mock(return_value=None)
        process_500 = Mock(return_value=None)
        app.route('/404', GET, Response.not_found)
        app.route('/500', GET, Response.internal_server_error)

        app.process_http_error_response(404)(process_404)
        app.process_http_error_response(500)(process_500)

        with app.requestcontext('/404'):
            app.view()
            assert process_500.call_count == 0
            assert process_404.call_count == 1

        with app.requestcontext('/500'):
            app.view()
            assert process_500.call_count == 1
            assert process_404.call_count == 1

    def test_process_teardown_called_on_teardown(self):
        app = FrescoApp()
        fn = Mock()
        app.process_teardown(fn)
        with app.requestcontext('/'):
            app.view()
            assert fn.call_count == 0
        assert fn.call_count == 1

    def test_adding_process_teardown_raises_error_after_app_starts(self):
        app = FrescoApp()
        with app.requestcontext('/') as c:
            iterator = app(c.request.environ, Mock())
            iterator.close()
        with pytest.raises(Exception):
            app.process_teardown(Mock())

    def test_exception_in_process_hooks_do_not_stop_processing(self):

        for hook, view in [('process_request', Response),
                           ('process_view', Response),
                           ('process_response', Response),
                           ('process_exception',
                            Mock(side_effect=Exception())),
                           ('process_http_error_response', Response.not_found),
                           ('process_teardown', Response)]:
            faulty_hook = Mock(side_effect=Exception())
            process_teardown = Mock(return_value=None)

            app = FrescoApp()
            app.route('/', GET, view)
            getattr(app, hook)(faulty_hook)
            app.process_teardown(process_teardown)

            with app.requestcontext('/'):
                app.view()

            assert faulty_hook.call_count == 1, hook
            assert process_teardown.call_count == 1


class TestProcessException(object):

    def exception_view(self):
        raise Exception()

    def iterator_exception_view(self):
        def content_iterator():
            yield b'fish'
            raise Exception()
            yield b'chips'
        return Response(content_iterator())

    def test_it_reraises_if_no_error_handlers_installed(self):
        app = FrescoApp()

        app.route('/', GET, self.exception_view)
        with app.requestcontext('/'):
            with pytest.raises(Exception):
                app.view()

    def test_it_reraises_if_no_500_error_handler_installed(self):

        app = FrescoApp()
        app.process_http_error_response(lambda req, res: None, 404)

        app.route('/', GET, self.exception_view)
        with app.requestcontext('/'):
            with pytest.raises(Exception):
                app.view()

    def test_it_calls_process_exception_handlers(self):
        app = FrescoApp()

        app.route('/', GET, self.exception_view)
        new_response = Response()
        process_exception = Mock(return_value=new_response)
        app.process_exception(process_exception)

        with app.requestcontext('/'):
            assert app.view() is new_response
            assert process_exception.call_count == 1

    def test_it_calls_process_exception_handlers_in_content_iteration(self):
        app = FrescoApp()

        app.route('/', GET, self.iterator_exception_view)
        process_exception = Mock(return_value=None)
        app.process_exception(process_exception)

        with app.requestcontext('/') as c:
            content_iterator = app(c.request.environ, Mock())
            assert process_exception.call_count == 0
            list(content_iterator)
            assert process_exception.call_count == 1

    def test_it_calls_error_handlers_in_middleware(self):
        """
        If a faulty middleware layer raises an exception it should
        trigger error handling.
        """
        def faulty_middleware(app):
            def middleware(environ, start_response):
                raise Exception()
            return middleware

        app = FrescoApp([Route('/', GET, lambda: Response("foo"))])
        app.add_middleware(faulty_middleware)
        process_exception = Mock(return_value=None)
        app.process_exception(process_exception)
        process_http_error_response = Mock(return_value=Response(b"bar"))
        app.process_http_error_response(process_http_error_response, 500)

        content = list(app(make_environ(), Mock()))
        assert process_exception.call_count == 1
        assert process_http_error_response.call_count == 1
        assert content == [b'bar']

    def test_it_associates_with_given_exception(self):

        def exception_view():
            raise ValueError()

        app = FrescoApp()
        app.route('/', GET, exception_view)
        process_TypeError = Mock(return_value=None)
        process_ValueError = Mock(return_value=None)
        app.process_exception(TypeError)(process_TypeError)
        app.process_exception(ValueError)(process_ValueError)

        with app.requestcontext('/'):
            app.view()
            assert process_TypeError.call_count == 0
            assert process_ValueError.call_count == 1

    def test_it_calls_http_error_handler(self):
        app = FrescoApp()

        app.route('/', GET, self.exception_view)
        new_response = Response()
        process_http_error_response = Mock(return_value=new_response)
        app.process_http_error_response(process_http_error_response)

        with app.requestcontext('/'):
            assert app.view() is new_response
            assert process_http_error_response.call_count == 1
            req, res = process_http_error_response.call_args[0]
            assert res.status_code == 500


class TestIncludeApp(object):

    def test_it_routes_to_an_included_app(self):

        app = FrescoApp()

        @app.route('/', GET)
        def view():
            return Response([b'ok'])

        app2 = FrescoApp()
        app2.include('/app1', app)

        with app2.requestcontext('/'):
            assert app2.view().status_code == 404

        with app2.requestcontext('/app1/'):
            assert list(app2.view().content_iterator) == [b'ok']

    def test_included_app_can_use_urlfor(self):

        def view():
            url = urlfor(view)
            return Response(url)

        app = FrescoApp()
        app.route('/', GET, view)
        app2 = FrescoApp()
        app2.include('/app1', app)

        with app2.requestcontext('/app1/'):
            assert list(app2.view().content_iterator) == \
                    [b'http://localhost/app1/']


class TestTrailingSlashes(object):
    """\
    The general principle is that if a GET or HEAD request is received for a
    URL without a trailing slash and no match is found, the app will look for a
    URL with a trailing slash, and redirect the client if such a route exists.
    """

    def test_no_trailing_slash(self):

        def foo():
            return Response(['foo'])

        app = FrescoApp()
        app.route('/foo/', GET, foo)

        with app.requestcontext('/foo'):
            assert app.view().status_code == 301
            assert app.view().get_header('location') == 'http://localhost/foo/'


class TestViewCollection(object):

    def test_appdef(self):

        app = FrescoApp()
        app.include('/', fixtures.CBV('bananas!'))
        with app.requestcontext('/'):
            assert list(app.view().content_iterator) == [b'bananas!']

    def test_appdef_url_generation(self):

        foo = fixtures.CBV('foo!')
        bar = fixtures.CBV('bar!')
        baz = fixtures.CBV('baz!')

        app = FrescoApp(views=foo)
        app.include('/bar', bar)
        app.include('/baz', baz)

        with app.requestcontext():
            assert urlfor(foo.index_html) == 'http://localhost/'
            assert urlfor(bar.index_html) == 'http://localhost/bar/'
            assert urlfor(baz.index_html) == 'http://localhost/baz/'

    def test_instance_available_in_context(self):

        s = []

        class MyCBV(fixtures.CBV):

            def index_html(self):
                from fresco import context
                s.append(context.view_self)
                return Response([])

        instance = MyCBV('foo!')
        app = FrescoApp(views=instance)

        with app.requestcontext('/'):
            app.view()
            assert s[0] is instance


class TestContextAttributes(object):

    def test_app_is_set(self):

        def check_app(expected):
            assert context.app is expected
            return Response([])

        app1 = FrescoApp()
        app2 = FrescoApp()

        app1.route('/', GET, check_app, {'expected': app1})
        app2.route('/', GET, check_app, {'expected': app2})

        with app1.requestcontext('/'):
            app1.view()

        with app2.requestcontext('/'):
            app2.view()


class TestAppRequestContext(object):

    def _middleware(self, app):
        def middleware(environ, start_response):
            environ['sausages'] = 1
            return app(environ, start_response)
        return middleware

    def test_creates_isolated_context(self):

        app = FrescoApp()
        with app.requestcontext():
            context.request = 'foo'

            with app.requestcontext():
                context.request = 'bar'
                assert context.request == 'bar'

            assert context.request == 'foo'

    def test_parses_full_url(self):

        with FrescoApp().requestcontext('https://arthur@example.org:123/?x=y'):
            assert context.request.environ['HTTPS'] == 'on'
            assert context.request.environ['REMOTE_USER'] == 'arthur'
            assert context.request.environ['HTTP_HOST'] == 'example.org:123'
            assert context.request.environ['SCRIPT_NAME'] == ''
            assert context.request.environ['PATH_INFO'] == '/'
            assert context.request.environ['QUERY_STRING'] == 'x=y'

    def test_it_posts_data(self):
        with FrescoApp().requestcontext_post(data={'foo': 'bar'}) as c:
            assert c.request.method == 'POST'
            assert list(c.request.form.allitems()) == [('foo', 'bar')]
            assert list(c.request.query.allitems()) == []

    def test_it_puts_data(self):
        with FrescoApp().requestcontext_put(data={'foo': 'bar'}) as c:
            assert c.request.method == 'PUT'
            assert list(c.request.form.allitems()) == [('foo', 'bar')]

    def test_it_posts_data_multipart(self):
        with FrescoApp().requestcontext_post(
                files=[('foo', 'foo.txt', 'text/plain', b'bar')]) as c:
            assert c.request.files['foo'].filename == 'foo.txt'
            assert c.request.files['foo'].file.read() == b'bar'

    def test_it_posts_raw_data(self):
        with FrescoApp().requestcontext_post(data=b'xyzzy',
                                             content_type='text/spell') as c:
            assert c.request.body == 'xyzzy'

    def test_it_converts_header_names(self):
        with FrescoApp().requestcontext(user_agent='foo'):
            assert context.request.environ['HTTP_USER_AGENT'] == 'foo'

    def test_it_converts_wsgi_keys(self):
        # CONTENT_TYPE is both a WSGI core environ key and a standard request
        # header. The WSGI key must win.
        with FrescoApp().requestcontext(content_type='foo'):
            assert context.request.environ['CONTENT_TYPE'] == 'foo'

    def test_it_invokes_middleware(self):

        app = FrescoApp()
        app.add_middleware(self._middleware)
        with app.requestcontext() as c:
            assert 'sausages' in c.request.environ

    def test_it_skips_middleware(self):

        app = FrescoApp()
        app.add_middleware(self._middleware)
        with app.requestcontext(middleware=False) as c:
            assert 'sausages' not in c.request.environ

    def test_it_closes_middleware(self):

        close = Mock()

        def middleware(app):
            def middleware(environ, start_response):
                return ClosingIterator(app(environ, start_response), close)
            return middleware

        app = FrescoApp()
        app.add_middleware(middleware)
        with app.requestcontext():
            pass
        assert close.call_count == 1

    def test_it_calls_first_iteration(self):
        """
        Some middleware waits until the first iteration to do things, so make
        sure we trigger this
        """
        from itertools import count
        counter = count()

        def middleware(app):
            def middleware(environ, start_response):
                iterator = app(environ, start_response)
                for item in iterator:
                    yield next(counter)
            return middleware

        app = FrescoApp()
        app.add_middleware(middleware)
        with app.requestcontext():
            assert next(counter) == 1

    def test_it_calls_late_added_middleware(self):

        calls = []

        def middleware(app):
            calls.append('middleware initialized')

            def middleware(env, start_response):
                calls.append('middleware called')
                return app(env, start_response)
            return middleware

        app = FrescoApp()
        app.route('/', GET, lambda: Response())

        # Put a request through app to initialize the WSGI stack
        with app.requestcontext('/'):
            app.view()

        app.add_middleware(middleware)

        # The request should cause the WSGI app to be rebuilt with the new
        # middleware included
        with app.requestcontext('/'):
            app.view()

        assert calls == ['middleware initialized', 'middleware called']


class TestUrlfor(object):

    def test_urlfor_on_aliased_functions(self):

        view = lambda: None
        setattr(fixtures, 'aliased_view', view)

        app = FrescoApp()
        app.route('/', GET, view)
        with app.requestcontext():
            assert urlfor(view) == 'http://localhost/'
            assert urlfor('fresco.tests.fixtures.aliased_view') == \
                    'http://localhost/'

        delattr(fixtures, 'aliased_view')

    def test_urlfor_with_view_function(self):

        def view():
            return Response(['ok'])

        app = FrescoApp()
        app.route('/foo', GET, view)
        with app.requestcontext():
            assert urlfor(view) == 'http://localhost/foo'

    def test_urlfor_allows_script_name(self):

        def view():
            return Response(['ok'])

        app = FrescoApp()
        app.route('/foo', GET, view)
        with app.requestcontext():
            assert urlfor(view, _script_name='/abc') ==\
                    'http://localhost/abc/foo'

    def test_urlfor_with_string(self):
        app = FrescoApp()
        app.route('/myviewfunc', GET, fixtures.module_level_function)
        with app.requestcontext():
            assert urlfor('fresco.tests.fixtures.module_level_function') ==\
                    'http://localhost/myviewfunc'

    def test_urlfor_drops_query(self):
        myviewfunc = lambda req: Response([])
        app = FrescoApp()
        app.route('/', GET, myviewfunc)
        with app.requestcontext():
            assert urlfor(myviewfunc) == 'http://localhost/'

    def test_urlfor_generates_first_route(self):

        myviewfunc = lambda req: Response([])
        app = FrescoApp()
        app.route('/1', GET, myviewfunc)
        app.route('/2', GET, myviewfunc)
        with app.requestcontext():
            assert urlfor(myviewfunc) == 'http://localhost/1'

    def test_urlfor_with_class_based_view_spec(self):

        app = FrescoApp()
        app.include('/foo', fixtures.CBV('x'))
        with app.requestcontext():
            assert urlfor('fresco.tests.fixtures.CBV.index_html') == \
                        'http://localhost/foo/'

    def test_it_uses_values_from_path_defaults(self):
        app = FrescoApp()
        app.route('/<test:int>', GET, lambda: None,
                  name='test', test_default=1)
        with app.requestcontext():
            assert urlfor('test') == 'http://localhost/1'

    def test_it_uses_callable_values_from_path_defaults(self):
        generate_value = Mock(return_value=1)
        app = FrescoApp()
        app.route('/<test:int>', GET, lambda: None,
                  name='test', test_default=generate_value)
        with app.requestcontext() as c:
            assert urlfor('test') == 'http://localhost/1'
            assert generate_value.call_args_list == [call(c.request)]

    def test_path_defaults_removed_from_view_kwargs(self):
        app = FrescoApp()
        view = Mock(return_value=Response())
        app.route('/<test:int>', GET, view,
                  name='test', test_default=1)
        with app.requestcontext('/2'):
            app.view()
            assert view.call_args_list == [call(test=2)]



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

import sys
from fresco import context

__all__ = 'FrescoCoreMiddleware', 'XForwarded'


class FrescoCoreMiddleware(object):

    def __init__(self, wsgiapp, frescoapp):
        self.wsgiapp = wsgiapp
        self.frescoapp = frescoapp

    def __call__(self, environ, start_response):

        request = self.frescoapp.request_class(environ)
        context.push(request=request, app=self.frescoapp)
        iterator = None
        try:
            iterator = self.wsgiapp(environ, start_response)
            for item in iterator:
                yield item
        except Exception:
            exc_info = sys.exc_info()
            request = context.request
            try:
                response = self.frescoapp.handle_exception(request)
                if '400' <= response.status <= '599':
                    response = self.frescoapp.handle_http_error_response(
                        request, response)

                exc_start_response = lambda s, h, exc_info=exc_info: \
                        start_response(s, h, exc_info)
                for item in response(environ, exc_start_response):
                    yield item
            finally:
                del exc_info
        finally:
            try:
                if self.frescoapp.process_teardown_handlers:
                    self.frescoapp.call_process_teardown_handlers(request)
            finally:
                try:
                    close = getattr(iterator, 'close', None)
                    if close is not None:
                        close()
                finally:
                    context.pop()


class XForwarded(object):
    """\
    Modify the WSGI environment so that the X_FORWARDED_* headers are observed
    and generated URIs are correct in a proxied environment.

    Use this whenever the WSGI application server is sitting behind
    Apache or another proxy server.

    It is easy for clients to spoof the X-Forwarded-For header. You can largely
    protect against this by listing all trusted proxy server addresses in
    ``trusted``. See http://en.wikipedia.org/wiki/X-Forwarded-For for more
    details.

    HTTP_X_FORWARDED_FOR is substituted for REMOTE_ADDR and
    HTTP_X_FORWARDED_HOST for SERVER_NAME. If HTTP_X_FORWARDED_SSL is set, then
    the wsgi.url_scheme is modified to ``https`` and ``HTTPS`` is set to
    ``on``.

    Example::

        >>> from fresco import FrescoApp, context, GET, Response
        >>> from flea import Agent
        >>> app = FrescoApp()
        >>> @app.route('/', GET)
        ... def view():
        ...     return Response(["URL is ", context.request.url,
        ...                      "; REMOTE_ADDR is ",
        ...                      context.request.remote_addr])
        ...
        >>> r = Agent(XForwarded(app))
        >>> response = r.get('/',
        ...     SERVER_NAME='wsgiserver-name',
        ...     SERVER_PORT='8080',
        ...     HTTP_HOST='wsgiserver-name:8080',
        ...     REMOTE_ADDR='127.0.0.1',
        ...     HTTP_X_FORWARDED_HOST='real-name:81',
        ...     HTTP_X_FORWARDED_FOR='1.2.3.4'
        ... )
        >>> response.body
        u'URL is http://real-name:81/; REMOTE_ADDR is 1.2.3.4'
        >>> response = r.get('/',
        ...     SERVER_NAME='wsgiserver-name',
        ...     SERVER_PORT='8080',
        ...     HTTP_HOST='wsgiserver-name:8080',
        ...     REMOTE_ADDR='127.0.0.1',
        ...     HTTP_X_FORWARDED_HOST='real-name:443',
        ...     HTTP_X_FORWARDED_FOR='1.2.3.4',
        ...     HTTP_X_FORWARDED_SSL='on'
        ... )
        >>> response.body
        u'URL is https://real-name/; REMOTE_ADDR is 1.2.3.4'
    """

    def __init__(self, app, trusted=None):
        self.app = app
        if trusted:
            self.trusted = set(trusted)
        else:
            self.trusted = set()

    def __call__(self, environ, start_response):
        """\
        Call the WSGI app, passing it a modified environ
        """
        is_ssl = environ.get('HTTP_X_FORWARDED_SSL') == 'on'

        if 'HTTP_X_FORWARDED_HOST' in environ:
            host = environ['HTTP_X_FORWARDED_HOST']

            if ':' in host:
                port = host.split(':')[1]
            else:
                port = is_ssl and '443' or '80'

            environ['HTTP_HOST'] = host
            environ['SERVER_PORT'] = port

        if is_ssl:
            environ['wsgi.url_scheme'] = 'https'
            environ['HTTPS'] = 'on'

        try:
            forwards = environ['HTTP_X_FORWARDED_FOR'].split(', ')\
                       + [environ.get('REMOTE_ADDR', '')]
        except KeyError:
            # No X-Forwarded-For header?
            return self.app(environ, start_response)

        if self.trusted:
            for ip in forwards[::-1]:
                # Find the first non-trusted ip; this is our remote address
                if ip not in self.trusted:
                    environ['REMOTE_ADDR'] = ip
                    break
        else:
            environ['REMOTE_ADDR'] = forwards[0]

        return self.app(environ, start_response)



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""
Security-related utilities
"""
from __future__ import absolute_import


def check_equal_constant_time(a, b):
    """
    Return ``True`` if string ``a`` is equal to string ``b``.

    If ``a`` and ``b`` are of the same length, this function will take the same
    amount of time to execute, regardless of whether or not a and b are equal.
    """
    if len(a) != len(b):
        return False
    result = 0
    for c1, c2 in zip(a, b):
        result |= ord(c1) ^ ord(c2)
    return result == 0



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from __future__ import absolute_import

try:
    from urllib.parse import quote, unquote
except ImportError:
    from urllib import quote, unquote  # NOQA
from datetime import datetime, timedelta
from calendar import timegm
from email.utils import formatdate
import warnings

from .compat import ustr


class RequestCookie(ustr):

    @property
    def value(self):
        warnings.warn("The value attribute on request cookies is deprecated",
                      DeprecationWarning, stacklevel=2)
        return self


class Cookie(object):
    """
    Represents an HTTP cookie.

    See rfc2109, HTTP State Management Mechanism

    Example:

        >>> from fresco.cookie import Cookie
        >>> c = Cookie('session_id', 'abc123')
        >>> c.path = '/cgi-bin'
        >>> c.domain = '.example.org'
        >>> c.path
        '/cgi-bin'
        >>> str(c)
        'session_id=abc123;Domain=.example.org;Path=/cgi-bin'
    """
    attributes = [
        ("Comment", "comment"),
        ("Domain", "domain"),
        ("Expires", "expires"),
        ("Max-Age", "max_age"),
        ("Path", "path"),
        ("Secure", "secure"),
        ("HttpOnly", "httponly"),
    ]

    bool_attributes = set(['secure', 'httponly'])

    def __init__(
        self, name, value, max_age=None, expires=None, path='/',
        secure=False, domain=None, comment=None, httponly=False,
        version=1
    ):
        """
        Initialize a ``Cookie`` instance.
        """
        self.name = name
        self.value = value
        self.max_age = max_age
        self.path = path
        self.secure = secure
        self.domain = domain
        self.comment = comment
        self.version = version
        self.expires = expires
        self.httponly = httponly

    def __str__(self):
        """
        Returns a string representation of the cookie in the format, eg
        ``session_id=abc123;Path=/cgi-bin;Domain=.example.com;HttpOnly``
        """
        cookie = ['%s=%s' % (self.name, quote(str(self.value)))]
        for cookie_name, att_name in self.attributes:
            value = getattr(self, att_name, None)
            if att_name in self.bool_attributes:
                if value:
                    cookie.append(cookie_name)
            elif value is not None:
                cookie.append('%s=%s' % (cookie_name, str(value)))
        return ';'.join(cookie)

    def set_expires(self, dt):
        """
        Set the cookie ``expires`` value to ``datetime`` object ``dt``
        """
        self._expires = dt

    def get_expires(self):
        """
        Return the cookie ``expires`` value as an instance of ``datetime``.
        """
        if self._expires is None and self.max_age is not None:
            if self.max_age == 0:
                # Make sure immediately expiring cookies get a date firmly in
                # the past.
                self._expires = datetime(1980, 1, 1)
            else:
                self._expires = datetime.now() + timedelta(seconds=self.max_age)

        if isinstance(self._expires, datetime):
            return formatdate(timegm(self._expires.utctimetuple()))

        else:
            return self._expires

    expires = property(get_expires, set_expires)


def expire_cookie(cookie_or_name, *args, **kwargs):
    """
    Synopsis:

        >>> from fresco import Response
        >>> from fresco.cookie import expire_cookie
        >>> def view():
        ...     return Response(set_cookie=expire_cookie('X', path='/'))
        ...
        >>> from fresco import FrescoApp
        >>> with FrescoApp().requestcontext() as c:
        ...     print(view().get_header('Set-Cookie'))
        ...
        X=;Expires=Tue, 01 Jan 1980 00:00:00 -0000;Max-Age=0;Path=/
    """
    if isinstance(cookie_or_name, Cookie):
        expire = cookie_or_name
    else:
        expire = Cookie(name=cookie_or_name, value='', *args, **kwargs)
    return Cookie(
        name=expire.name,
        value='',
        expires=datetime(1980, 1, 1),
        max_age=0,
        domain=kwargs.get('domain', expire.domain),
        path=kwargs.get('path', expire.path)
    )


def parse_cookie_header(cookie_string, unquote=unquote):
    """
    Return a list of cookie (name, value) pairs read from the request headers.

    :param cookie_string: The cookie, eg ``CUSTOMER=FRED``
    :param unquote: A function to decode values. By default values are assumed
                    to be url quoted. If ``None`` the raw value will be
                    returned
    """
    if not cookie_string:
        return []
    cookies = []

    for part in cookie_string.split(";"):

        try:
            k, v = part.strip().split("=", 1)
        except ValueError:
            continue

        if k[0] == '$':
            # An attribute (eg path or domain) pertaining to the most recently
            # read cookie.
            # These were defined in RFC2109 and RFC2965, but removed in RFC6265
            # (April 2011) and as far as I know not used by browsers.
            # In any case we are only interested in parsing the value,
            # so we can drop these.
            continue

        # Unquote quoted values ('"..."' => '...')
        if v and '"' == v[0] == v[-1] and len(v) > 1:
            v = v[1:-1]

        if unquote:
            k, v = unquote(k), unquote(v)
        cookies.append((k, RequestCookie(v)))

    return cookies
